#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <sil.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include "sample1.h"

//
// WARNING 1: at the moment you need to apply a hard reset to the board
// for this example to work. Soft reset is not enough (probably there
// is some register I have to set up)
//
// WARNING 2: this code contains GPL code so the resulting binary
// will have GPL license, instead of TOPPERS license.
//

/*
 *  サービスコールのエラーのログ出力
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
        if (ercd < 0) {
                t_perror(LOG_ERROR, file, line, expr, ercd);
        }
}

#define SVC_PERROR(expr)        svc_perror(__FILE__, __LINE__, #expr, (expr))

/*
 * GPIO driver (linux/drivers/gpio/xilinx_gpiops.c)
 *
 * Xilinx PS GPIO device driver
 *
 * 2009-2011 (c) Xilinx, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 *
 */
#define ZYNQ_GPIO_BASE   ((void *)0xE000A000)

#define XGPIOPS_DATA_LSW_OFFSET(BANK)   (0x000 + (8 * BANK))    /* LSW Mask & Data -WO */
#define XGPIOPS_DATA_MSW_OFFSET(BANK)   (0x004 + (8 * BANK))    /* MSW Mask & Data -WO */
#define XGPIOPS_DIRM_OFFSET(BANK)       (0x204 + (0x40 * BANK)) /* Direction mode reg-RW */
#define XGPIOPS_OUTEN_OFFSET(BANK)      (0x208 + (0x40 * BANK)) /* Output enable reg-RW */

/* Read/Write access to the GPIO PS registers */

static inline uint32_t __raw_readl(const volatile void *addr)
{
        return *(const volatile uint32_t *) addr;
}

static inline void __raw_writel(uint32_t b, volatile void *addr)
{
        *(volatile uint32_t *) addr = b;
}

#define xgpiops_readreg(offset)         __raw_readl(ZYNQ_GPIO_BASE + (offset))
#define xgpiops_writereg(val, offset)   __raw_writel(val, ZYNQ_GPIO_BASE + (offset))

static unsigned int xgpiops_pin_table[] = {
        31, /* 0 - 31 */
        53, /* 32 - 53 */
        85, /* 54 - 85 (61-69 are LEDs) */
        117 /* 86 - 117 */
};

/**
 * xgpiops_get_bank_pin - Get the bank number and pin number within that bank
 * for a given pin in the GPIO device
 * @pin_num:    gpio pin number within the device
 * @bank_num:   an output parameter used to return the bank number of the gpio
 *              pin
 * @bank_pin_num: an output parameter used to return pin number within a bank
 *                for the given gpio pin
 *
 * Returns the bank number.
 */
static inline void xgpiops_get_bank_pin(unsigned int pin_num,
                                        unsigned int *bank_num,
                                        unsigned int *bank_pin_num)
{
        for (*bank_num = 0; *bank_num < 4; (*bank_num)++)
                if (pin_num <= xgpiops_pin_table[*bank_num]) break;

        if (*bank_num == 0)
                *bank_pin_num = pin_num;
        else
                *bank_pin_num = pin_num %
                (xgpiops_pin_table[*bank_num - 1] + 1);
}

/**
 * xgpiops_set_value - Modify the state of the pin with specified value
 * @chip:       gpio_chip instance to be worked on
 * @pin:        gpio pin number within the device
 * @state:      value used to modify the state of the specified pin
 *
 * This function calculates the register offset (i.e to lower 16 bits or
 * upper 16 bits) based on the given pin number and sets the state of a
 * gpio pin to the specified value. The state is either 0 or non-zero.
 */
static void xgpiops_set_value(unsigned int pin, int state)
{
        unsigned int reg_offset;
        unsigned int bank_num, bank_pin_num;

        xgpiops_get_bank_pin(pin, &bank_num, &bank_pin_num);

        if (bank_pin_num >= 16) {
                bank_pin_num -= 16; /* only 16 data bits in bit maskable reg */
                reg_offset = XGPIOPS_DATA_MSW_OFFSET(bank_num);
        } else
                reg_offset = XGPIOPS_DATA_LSW_OFFSET(bank_num);

        /*
         * get the 32 bit value to be written to the mask/data register where
         * the upper 16 bits is the mask and lower 16 bits is the data
         */
        if (state)
                state = 1;
        state = ~(1 << (bank_pin_num + 16)) & ((state << bank_pin_num) | 0xFFFF0000);

        loc_spn(GPIO_LOCK);
        xgpiops_writereg(state, reg_offset);
        unl_spn(GPIO_LOCK);
}

/**
 * xgpiops_dir_out - Set the direction of the specified GPIO pin as output
 * @chip:       gpio_chip instance to be worked on
 * @pin:        gpio pin number within the device
 * @state:      value to be written to specified pin
 *
 * This function sets the direction of specified GPIO pin as output, configures
 * the Output Enable register for the pin and uses xgpiops_set to set the state
 * of the pin to the value specified. Returns 0 always.
 */
static int xgpiops_dir_out(unsigned int pin, int state)
{
        unsigned int reg, bank_num, bank_pin_num;

        xgpiops_get_bank_pin(pin, &bank_num, &bank_pin_num);

        loc_spn(GPIO_LOCK);

        /* set the GPIO pin as output */
        reg = xgpiops_readreg(XGPIOPS_DIRM_OFFSET(bank_num));
        reg |= 1 << bank_pin_num;
        xgpiops_writereg(reg, XGPIOPS_DIRM_OFFSET(bank_num));

        /* configure the output enable reg for the pin */
        reg = xgpiops_readreg(XGPIOPS_OUTEN_OFFSET(bank_num));
        reg |= 1 << bank_pin_num;
        xgpiops_writereg(reg, XGPIOPS_OUTEN_OFFSET(bank_num));

        /* set the state of the pin */
        xgpiops_set_value(pin, state);

        unl_spn(GPIO_LOCK);

        return 0;
}

/*
 * LED loop
 */
static void led_loop(unsigned int pin)
{
        xgpiops_dir_out(pin, 1);

        while(1) {
                xgpiops_set_value(pin, 1);
                dly_tsk(1000);
                xgpiops_set_value(pin, 0);
                dly_tsk(1000);
        }
}

/*
 * MAIN TASK Core 1
 */
void main_task_1(intptr_t exinf)
{
        led_loop(61);
}

/*
 * MAIN TASK Core 2
 */
void main_task_2(intptr_t exinf)
{
        led_loop(62);
}
