/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 *
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2006-2010 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *      Copyright (C) 2007-2010 by Industrial Technology Institute,
 *              Miyagi Prefectural Government, JAPAN
 *
 *      上記著作権者は，以下の(1)〜(4)の条件を満たす場合に限り，本ソフトウェ
 *      ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *      変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *      (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *              権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *              スコード中に含まれていること．
 *      (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *              用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *              者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *              の無保証規定を掲載すること．
 *      (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *              用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *              と．
 *        (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *                作権表示，この利用条件および下記の無保証規定を掲載すること．
 *        (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *                報告すること．
 *      (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *              害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *              また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *              由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *              免責すること．
 *
 *      本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *      よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *      に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *      アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *      の責任を負わない．
 *
 *  ZYNQ porting
 *  2012 (c) Daniel Sangorrin (daniel.sangorrin@gmail.com)
 *
 *  This file contains calls to the Xilinx UART driver (xuartps.c)
 *  in order to implement the FMP serial input/output (SIO) interface.
 */
#include <kernel.h>
#include <t_syslog.h>
#include "target_serial.h"
#include "target_syssvc.h"

/*
 * Configuration constants (comment out if necessary)
 */
// #define XUARTPS_POLLING_MODE /* use polling instead of interrupts */

/*
 * Module variables
 */
static intptr_t the_exinf; /* copy of the serial module exinf */

/*
 * UART Registers
 */
#define XUARTPS_CR_REG                  0xE0001000  /* Control Register */
#define     XUARTPS_CR_STOPBRK          0x00000100  /* Stop TX break */
#define     XUARTPS_CR_TX_EN            0x00000010  /* TX enabled */
#define     XUARTPS_CR_RX_EN            0x00000004  /* RX enabled */
#define     XUARTPS_CR_TX_DIS           0x00000020  /* TX disabled. */
#define     XUARTPS_CR_RX_DIS           0x00000008  /* RX disabled. */
#define     XUARTPS_CR_TXRST            0x00000002  /* TX logic reset */
#define     XUARTPS_CR_RXRST            0x00000001  /* RX logic reset */

#define XUARTPS_MR_REG                  0xE0001004  /* Mode Register */
#define     XUARTPS_MR_STOPMODE_1_BIT   0x00000000  /* 1 stop bit */
#define     XUARTPS_MR_PARITY_NONE      0x00000020  /* No parity mode */
#define     XUARTPS_MR_CHARLEN_8_BIT    0x00000000  /* 8 bits data */
#define     XUARTPS_MR_CLKSEL           0x00000001  /* Input clock select */
#define     XUARTPS_MR_CCLK             0x00000400  /* Input clock select */

#define XUARTPS_IER_REG                 0xE0001008  /* Interrupt Enable */
#define XUARTPS_IDR_REG                 0xE000100C  /* Interrupt Disable */
#define XUARTPS_ISR_REG                 0xE0001014  /* Interrupt Status */
#define    XUARTPS_IXR_TXEMPTY          0x00000008 /* TX FIFO empty int */
#define    XUARTPS_IXR_RXTRIG           0x00000001 /* RX FIFO trigger int */

#define XUARTPS_BAUDGEN_REG             0xE0001018  /* Baud Rate Generator */
#define XUARTPS_BAUDDIV_REG             0xE0001034  /* Baud Rate Divider */

#define XUARTPS_RXTOUT_REG              0xE000101C  /* RX Timeout [7:0] */

#define XUARTPS_SR_REG                  0xE000102C  /* Channel Status */
#define         XUARTPS_SR_TXFULL       0x00000010 /* TX FIFO full */
#define         XUARTPS_SR_RXEMPTY      0x00000002 /* RX FIFO empty */

#define XUARTPS_FIFO_REG                0xE0001030  /* FIFO */
#define XUARTPS_RXWM_REG                0xE0001020  /* RX FIFO Trigger */

#define REG(x) *((volatile uint32_t *)(x))

/*
 * Definition of the 115200bps baud rate constants
 *        rate = clk / (CD * (BDIV + 1))
 *         clk = MR.CLKSEL? inclk : inclk/8  (e.g. inclk)
 *       inclk = MR.CCLK? uart_clock : apb clock (e.g. uart_clock=50MHz)
 */
#define UART_BAUD_115K          0x56  /* 0x56 (50MHz) or 0x11 (33.3MHz) */
#define UART_BAUDDIV_115K       0x4   /* 0x4 (50MHz) or 0x6 (33.3MHz) */

/*
 * AUXILIAR FUNCTIONS
 */

/*
 *  Is there a received character?
 */
Inline bool_t
xuartps_getready()
{
        return (REG(XUARTPS_SR_REG) & XUARTPS_SR_RXEMPTY) == 0;
}

/*
 *  Can I send a new character?
 */
Inline bool_t
xuartps_putready()
{
        return (REG(XUARTPS_SR_REG) & XUARTPS_SR_TXFULL) == 0;
}

/*
 *  Get character from buffer
 */
Inline char_t
xuartps_getchar()
{
        return (char_t)REG(XUARTPS_FIFO_REG);
}

/*
 * Write character to buffer
 */
Inline void
xuartps_putchar(char_t c)
{
        REG(XUARTPS_FIFO_REG) = (uint32_t)c;
}


#ifndef XUARTPS_POLLING_MODE
/*
 *  Enable send interrupts
 */
Inline void
xuartps_enable_send()
{
        REG(XUARTPS_IER_REG) = XUARTPS_IXR_TXEMPTY;
}

/*
 *  Diable send interrupts
 */
Inline void
xuartps_disable_send()
{
        REG(XUARTPS_IDR_REG) = XUARTPS_IXR_TXEMPTY;
}

/*
 *  Enable receive interrupts
 */
Inline void
xuartps_enable_rcv()
{
        REG(XUARTPS_IER_REG) = XUARTPS_IXR_RXTRIG;
}

/*
 *  Disable receive interrupts
 */
Inline void
xuartps_disable_rcv()
{
        REG(XUARTPS_IDR_REG) = XUARTPS_IXR_RXTRIG;
}
#endif /* XUARTPS_POLLING_MODE */

/*
 * EARLY USAGE FUNCTIONS
 */

/*
 *  Low-level polling
 */
void
xuartps_pol_putc(char_t c)
{
        while (!xuartps_putready());
        xuartps_putchar(c);
}

/*
 *  Initialize the UART
 */
void
xuartps_init()
{
        /* Disable all interrupts */
        REG(XUARTPS_IDR_REG) = 0x00001FFF;

        /* Clear pending interrupts */
        REG(XUARTPS_ISR_REG) = REG(XUARTPS_ISR_REG);

        /* Reset the UART and disable TX/RX*/
        REG(XUARTPS_CR_REG) = XUARTPS_CR_TX_DIS |  XUARTPS_CR_RX_DIS |
                              XUARTPS_CR_TXRST  | XUARTPS_CR_RXRST;

        /* Set 115200bps baud rate */
        REG(XUARTPS_BAUDGEN_REG) = UART_BAUD_115K;
        REG(XUARTPS_BAUDDIV_REG) = UART_BAUDDIV_115K;

        // Set 8n1 (8 bits - No parity - 1 Stop bit)
        REG(XUARTPS_MR_REG) = XUARTPS_MR_CHARLEN_8_BIT |
                              XUARTPS_MR_PARITY_NONE   |
                              XUARTPS_MR_STOPMODE_1_BIT;

        // Set Rx trigger (1 byte)
        REG(XUARTPS_RXWM_REG)   = 0x01;

        // Set timeout (Nx4bits-time)
        REG(XUARTPS_RXTOUT_REG) = 10;

        // Enable transmission/reception
        REG(XUARTPS_CR_REG) = XUARTPS_CR_TX_EN | XUARTPS_CR_RX_EN |
                              XUARTPS_CR_STOPBRK;
}

/*
 * SIO INTERFACE IMPLEMENTATION
 */

/*
 * Open the serial port
 */
SIOPCB *
sio_opn_por(ID siopid, intptr_t exinf)
{
        ER ercd;

        the_exinf = exinf;

        ercd = ena_int(INTNO_SIO);
        assert(ercd == E_OK);
        return 0;
}

/*
 * Close the serial port
 */
void
sio_cls_por(SIOPCB *p_siopcb)
{
        ER ercd;

        // Disable transmission/reception
        REG(XUARTPS_CR_REG)  = XUARTPS_CR_STOPBRK | XUARTPS_CR_TX_DIS  |
                               XUARTPS_CR_RX_DIS;
        // Disable all interrupts
        REG(XUARTPS_IDR_REG) = 0x00001FFF;
        ercd = dis_int(INTNO_SIO);
        assert(ercd == E_OK);
}

/*
 *  SIO interrupt service routine (used in target_serial.cfg)
 */
void
sio_isr(intptr_t siopid)
{
#ifndef XUARTPS_POLLING_MODE
        /*
         * Perform callbacks to the serial module if a transmission
         * finished or some characters were received
         */
        if (xuartps_putready())
                sio_irdy_snd(the_exinf);

        if (xuartps_getready())
                sio_irdy_rcv(the_exinf);

        /* clear pending interrupts */
        REG(XUARTPS_ISR_REG) = REG(XUARTPS_ISR_REG);
#endif /* XUARTPS_POLLING_MODE */
}

/*
 * Send a character through the serial port
 */
bool_t
sio_snd_chr(SIOPCB *siopcb, char_t c)
{
#ifndef XUARTPS_POLLING_MODE
        if (!xuartps_putready()) return false;
#else
        while(!xuartps_putready());
#endif /* XUARTPS_POLLING_MODE */

        xuartps_putchar(c);
        return true;
}

/*
 * Receive a character through the serial port
 */
int_t
sio_rcv_chr(SIOPCB *siopcb)
{
#ifndef XUARTPS_POLLING_MODE
        if (!xuartps_getready()) return -1;
#else
        while(!xuartps_getready());
#endif /* XUARTPS_POLLING_MODE */

        return (int_t)(uint8_t)xuartps_getchar();
}

/*
 *  Enable callbacks from the serial port hardware when a character is
 *  received (cbrtn=SIO_RDY_RCV) or sent (cbrtn==SIO_RDY_SND).
 */
void
sio_ena_cbr(SIOPCB *siopcb, uint_t cbrtn)
{
#ifndef XUARTPS_POLLING_MODE
        if (cbrtn == SIO_RDY_SND)
                xuartps_enable_send();
        else if (cbrtn == SIO_RDY_RCV)
                xuartps_enable_rcv();
#endif /* XUARTPS_POLLING_MODE */
}

/*
 *  Disable callbacks from the serial port hardware.
 */
void
sio_dis_cbr(SIOPCB *siopcb, uint_t cbrtn)
{
#ifndef XUARTPS_POLLING_MODE
        if (cbrtn == SIO_RDY_SND)
                xuartps_disable_send();
        else if (cbrtn == SIO_RDY_RCV)
                xuartps_disable_rcv();
#endif /* XUARTPS_POLLING_MODE */
}
